----------------------------------------------------------------------
-- Galois Field Chapter 8 package body (GF2m.adb)
-- 
-- Defines useful funtions for GF (2^m)
-- of Chapter 8 examples
---------------------------------------------------------------------
package body Gf2m is   

   function Product_Alpha_A(A, F: Poly_Vector) return Poly_Vector is
      C: Poly_Vector;
   begin
      C(0) := M2and(A(M-1),F(0));
      for I in 1 .. M-1 loop C(I) := M2xor(A(I-1),M2and(A(M-1),F(I))); end loop;
      return C;
   end Product_Alpha_A; 

   ---------------------------------------------------------------------------------

   function Product_Alpha_Am(A, F: Poly_Vectorm) return Poly_Vectorm is
      C: Poly_Vectorm;
   begin
      C(0) := M2and(A(M),F(0));
      for I in 1 .. M loop C(I) := M2xor(A(I-1),M2and(A(M),F(I))); end loop;
      return C;
   end Product_Alpha_Am; 

   ---------------------------------------------------------------------------------

   function Product_Alpha(F: Poly_Vector) return Poly_Vector is
      C: Poly_Vector;
   begin
      C(0) := M2and(F(M-1),F(0));
      for I in 1 .. M-1 loop C(I) := M2xor(F(I-1),M2and(F(M-1),F(I))); end loop;
      return C;
   end Product_Alpha; 

   ---------------------------------------------------------------------------------
   
   function M2abv(X: Bit; Y: Poly_Vector) return Poly_Vector is
      Z : Poly_Vector;
   begin
      for I in 0 .. M-1 loop
         Z(I) := M2and(X,Y(I));
      end loop;
      return Z;
   end M2abv;
   
   ---------------------------------------------------------------------------------
   
   function M2xvv(X: Poly_Vector; Y: Poly_Vector) return Poly_Vector is
      Z : Poly_Vector;
   begin
      for I in 0 .. M-1 loop
         Z(I) := M2xor(X(I),Y(I));
      end loop;
      return Z;
   end M2xvv;

   ---------------------------------------------------------------------------------
   
   function M2xvvm(X: Poly_Vectorm; Y: Poly_Vectorm) return Poly_Vectorm is
      Z : Poly_Vectorm;
   begin
      for I in 0 .. M loop
         Z(I) := M2xor(X(I),Y(I));
      end loop;
      return Z;
   end M2xvvm;

   ---------------------------------------------------------------------------------

   function M2xvv2(X: Poly2_Vector; Y: Poly_Vector) return Poly2_Vector is
      Z : Poly2_Vector;
   begin
      for I in 0 .. M-1 loop
         Z(I) := M2xor(X(I),Y(I));
      end loop;
      for I in M .. 2*M-2 loop
         Z(I) := X(I);
      end loop;
      return Z;
   end M2xvv2;
   
   ---------------------------------------------------------------------------------

   function Lshift(X: Poly_Vector) return Poly_Vector is
      Z: Poly_Vector;
   begin
      for I in 0 .. M-2 loop
         Z(I) := X(I+1);
      end loop;
      Z(M-1) := 0;
      return Z;
   end Lshift;

   --------------------------------------------------------------------------------- 

   function Rshift(X: Poly_Vector) return Poly_Vector is
      Z: Poly_Vector;
   begin
      for I in 1 .. M-1 loop
         Z(I) := X(I-1);
      end loop;
      Z(0) := 0;
      return Z;
   end Rshift;

   --------------------------------------------------------------------------------- 

   function Lshiftm(X: Poly_Vectorm) return Poly_Vectorm is
      Z: Poly_Vectorm;
   begin
      for I in 0 .. M-1 loop
         Z(I) := X(I+1);
      end loop;
      Z(M) := 0;
      return Z;
   end Lshiftm;

   --------------------------------------------------------------------------------- 

   function Rshiftm(X: Poly_Vectorm) return Poly_Vectorm is
      Z: Poly_Vectorm;
   begin
      for I in 1 .. M loop
         Z(I) := X(I-1);
      end loop;
      Z(0) := 0;
      return Z;
   end Rshiftm;

   --------------------------------------------------------------------------------- 

   function Lshift2(X: Poly2_Vector) return Poly2_Vector is
      Z: Poly2_Vector;
   begin
      for I in 0 .. 2*M-3 loop
         Z(I) := X(I+1);
      end loop;
      Z(2*M-2) := 0;
      return Z;
   end Lshift2;

   --------------------------------------------------------------------------------- 

   function Poly_Multiplication(A,B: Poly_Vector) return Poly2_Vector is
      C: Poly2_Vector;
   begin
      for I in 0 .. 2*M-2 loop C(I) := 0; end loop;

      for K in 0 .. M-1 loop
         for I in 0 .. K loop
            C(K) := M2xor(C(K),M2and(A(I),B(K-I)));
         end loop;
      end loop;
   
      for K in M .. 2*M-2 loop
         for I in K .. 2*M-2 loop
            C(K) := M2xor(C(K),M2and(A(K-I+(M-1)),B(I-(M-1))));
         end loop;
      end loop;
   
      return C;
      
   end Poly_Multiplication;

   ---------------------------------------------------------------------------------

   function Reduction_Matrix_R (F: Poly_Vector) return Poly_Matrix_M1m2 is
      R: Poly_Matrix_M1m2;
   begin
      for J in 0 .. M-1 loop 
         for I in 0 .. M-2 loop
            R(J,I) := 0; 
         end loop;
      end loop;
      
      for J in 0 .. M-1 loop
         R(J,0) := F(J);
      end loop;
 
      for I in 1 .. M-2 loop
         for J in 0 .. M-1 loop
            if J = 0 then 
               R(J,I) := M2and(R(M-1,I-1),R(J,0));
            else
               R(J,I) := M2xor(R(J-1,I-1),M2and(R(M-1,I-1),R(J,0))); 
            end if;
         end loop;
      end loop;
      
      return R;
      
   end Reduction_Matrix_R;

   ---------------------------------------------------------------------------------

   function Matrix_P (F: Poly_Vector) return Poly_Matrix_M2m1 is
      P: Poly_Matrix_M2m1;
   begin
      for J in 0 .. M-1 loop
         P(0,J) := F(J);
      end loop;
   
      for I in 1 .. M-2 loop 
         for J in 0 .. M-1 loop
            P(I,J) := 0; 
         end loop;
      end loop;
    
      for I in 1 .. M-2 loop
         for J in 0 .. M-1 loop
            if J = 0 then 
               P(I,J) := P(I-1,M-1);
            else
               P(I,J) := M2xor(P(I-1,J-1),M2and(P(I-1,M-1),P(0,J))); 
            end if;
         end loop;
      end loop;
      
      return P;
      
   end Matrix_P;
   
   ---------------------------------------------------------------------------------

   function Mastrovito_Matrix (A: Poly_Vector; P: Poly_Matrix_M2m1) return Poly_Matrix is
      Z: Poly_Matrix;
   begin
      for I in 0 .. M-1 loop
         Z(I,0) := A(I);
      end loop;
   
      for I in 0 .. M-1 loop 
         for J in 1 .. M-1 loop
            Z(I,J) := 0; 
         end loop;
      end loop;
    
      for I in 0 .. M-1 loop
         for J in 1 .. M-1 loop
            for T in 0 .. J-1 loop
               Z(I,J) := M2xor(Z(I,J),M2and(P(J-1-T,I),A(M-1-T)));
            end loop;
            if I >= J then   
               Z(I,J) := M2xor(A(I-J),Z(I,J));        
            end if;
         end loop;
      end loop;
      
      return Z;
      
   end Mastrovito_Matrix;
   
   --------------------------------------------------------------------------------- 

   function Vector_D(A,B: Poly_Vector) return Poly_Vector is
      D: Poly_Vector;
   begin
      for I in 0 .. M-1 loop D(I) := 0; end loop;

      for I in 0 .. M-1 loop
         for K in 0 .. I loop
            D(I) := M2xor(D(I),M2and(A(K),B(I-K)));
         end loop;
      end loop;
   
      return D;
      
   end Vector_D;

   
   --------------------------------------------------------------------------------- 

   function Vector_E(A,B: Poly_Vector) return Poly_Vector2 is
      E: Poly_Vector2;
   begin
      for I in 0 .. M-2 loop E(I) := 0; end loop;

      for I in 0 .. M-2 loop
         for K in I .. M-2 loop
            E(I) := M2xor(E(I),M2and(A(M-1-(K-I)),B(K+1)));
         end loop;
      end loop;
   
      return E;
      
   end Vector_E;

   --------------------------------------------------------------------------------- 

   function Lsbfirst_Squarer(A,F: Poly_Vector) return Poly_Vector is
      W,B,Faux, Baux: Poly_Vector;

   begin
      for I in 0 .. M-1 loop W(I) := 0; Baux(I) := 0; end loop;
      for I in 0 .. (M-2)/2 loop W(2*I) := A(I); end loop; -- even m
   
      B := F; -- even m
      Faux := Product_Alpha(F);
   
      for K in 1 .. (M/2)-1 loop
         for I in 2 .. M-1 loop
            Baux(I) := M2xor(B(I-2),M2xor(M2and(B(M-1),Faux(I)),M2and(B(M-2),F(I))));
         end loop;
         Baux(1) := M2xor(M2and(B(M-1),Faux(1)),M2and(B(M-2),F(1)));
         Baux(0) := M2xor(M2and(B(M-1),Faux(0)),M2and(B(M-2),F(0)));
    
         W := M2xvv(W,M2abv(A((M/2)+K-1),B));
         B := Baux;
      end loop;   

      W := M2xvv(W,M2abv(A(M-1),B));

      return W;      

   end Lsbfirst_Squarer;

   --------------------------------------------------------------------------------- 

   function Lsbfirst(A,B,F: Poly_Vector) return Poly_Vector is
      Aux, C: Poly_Vector;

   begin
      for I in 0 .. M-1 loop C(I) := 0; end loop;
      Aux := A;   
      for I in 0 .. M-1 loop
         C := M2xvv(M2abv(B(I),Aux),C);
         Aux := Product_Alpha_A(Aux,F);   
      end loop;   

      return C;
   
   end Lsbfirst;

   --------------------------------------------------------------------------------- 

   function Bsquarer_Montgomery(A,F: Poly_Vector) return Poly_Vector is
      C: Poly2_Vector;
      D: Poly_Vector;

   begin
      for I in 0 .. 2*M-2 loop C(I) := 0; end loop;
      for I in 0 .. M-1 loop C(2*I) := A(I); end loop;

      for I in 0 .. M-1 loop
         if C(0) = 1 then 
            C := M2xvv2(C,F);
            C(M) := M2xor(C(M),1);
         end if;
         C := Lshift2(C);   
      end loop;   

      for I in 0 .. M-1 loop
         D(I) := C(I);
      end loop;

      return D;      

   end Bsquarer_Montgomery;

   --------------------------------------------------------------------------------- 

   function Bmult_Montgomery(A,B,F: Poly_Vector) return Poly_Vector is
      C: Poly_Vector;

   begin
      for I in 0 .. M-1 loop C(I) := 0; end loop;

      for I in 0 .. M-1 loop
         C := M2xvv(C,M2abv(A(I),B));
         if C(0) = 1 then 
            C := M2xvv(C,M2abv(C(0),F));
            C := Lshift(C);
            C(M-1) := 1;
         else
            C := Lshift(C);   
         end if;   
      end loop;   
      
      return C;
      
   end Bmult_Montgomery;

   --------------------------------------------------------------------------------- 

   function Unitym(A: Poly_Vectorm) return Integer is
      One: Integer;
   begin
      One := 0;
      if A(0) = 1 then One := 1; end if;
      for I in 1 .. M loop
         if A(I) = 1 then One := 0; end if;
      end loop;

      return One;
      
   end Unitym;
   
   --------------------------------------------------------------------------------- 

   function Degreem(A: Poly_Vectorm) return Integer is
      D: Integer;   
   
   begin
      D := 0;
      for I in 0 .. M loop
         if A(I) = 1 then 
            D := I; 
         end if;
      end loop;
      
      return D;
      
   end Degreem;
   
   --------------------------------------------------------------------------------- 

   function Nb_Sq(A: Poly_Vector) return Poly_Vector is
      B: Poly_Vector;
      A0: Bit;   
   
   begin
   
      for I in 0 .. M-1 loop
         B(I) := 0;
      end loop;
      
      A0 := A(M-1); 
      B := Rshift(A);
      B(0) := A0; 
   
      return B;
      
   end Nb_Sq;

   ----------------------------------------------------------------------------------

   function H_Function_Gf2_4(A,B: Poly_Vector) return Bit is
      C: Bit;
   begin
      C := M2xor(M2and(A(2),B(2)),M2xor(M2and(A(3),B(2)),M2xor(M2and(A(2),B(3)),M2xor(M2and(A(3),B(1)),M2xor(M2and(A(1),B(3)),M2xor(M2and(A(3),B(0)),M2xor(M2and(A(0),B(3)),M2xor(M2and(A(1),B(0)),M2and(A(0),B(1))))))))));
   
      return C;   
   
   end H_Function_Gf2_4;
   
   ----------------------------------------------------------------------------------

   function Nb_Multiplier(A,B: Poly_Vector; H: H_Array; W: W_Array) return Poly_Vector is
      C: Poly_Vector;
      R,T: Poly_Vector;
      Yij: Yij_Array;
      V,S,Te: Integer;
   
   begin
  
      V := M/2;
     
      for I in 0 .. M-1 loop
         for J in 1 .. M/2 loop
            Yij(I,J) := 0;
         end loop;
      end loop;
    
      for I in 0 .. M-1 loop
         R(I) := 0;
         T(I) := 0;
      end loop; 
      for I in 0 .. M-1 loop
         for J in 1 .. V loop
            Yij(I,J) := M2and(M2xor(A(I),A((I+J) mod M)),M2xor(B(I),B((I+J) mod M)));
         end loop;
      end loop;
      for I in 0 .. M-1 loop
         C(I) := M2and(A(I),B(I));
      end loop;
      for J in 1 .. V-1 loop
         for I in 0 .. M-1 loop
            T(I) := 0;
         end loop;
         for K in 1 .. H(J) loop
            for I in 0 .. M-1 loop
               R(I) := Yij((I-W(J,K)) mod M,J);
            end loop;
            T := M2xvv(T,R);
         end loop;
         C := M2xvv(C,T);
      end loop;
      for I in 0 .. M-1 loop
         T(I) := 0;
      end loop;
      if (M rem 2) /= 0 then
         S := H(V);
         Te := M;
      else
         S := H(V)/2;
         Te := M/2;
      end if;      
      for I in 0 .. Te-1 loop
         Yij(I,V) := M2and(M2xor(A(I),A((V+I) mod M)),M2xor(B(I),B((V+I) mod M)));
      end loop;
      if (M rem 2) = 0 then
         for I in 0 .. (M/2)-1 loop 
            Yij(I+V,V) := Yij(I,V);
         end loop;
      end if;   
      for K in 1 .. S loop
         for I in 0 .. Te-1 loop
            R(I) := Yij((I-W(V,K)) mod M,V);
         end loop;
         if (M rem 2) = 0 then
            for I in 0 .. (M/2)-1 loop
               R(I+M/2) := R(I);
            end loop;        
         end if;
         T := M2xvv(T,R);
      end loop; 
   
      C:= M2xvv(C,T);       
           
      return C;

   end Nb_Multiplier;

   -------------------------------------------------------------------

   function Log(X: Integer) return Integer is
      Y: Integer;
      K: Integer;
   begin
      K := 0;
      Y := 1;
      while K = 0 loop
         if 2**Y <= X then
            Y := Y + 1;
         else 
            K := 1;
         end if;
      end loop;
      return (Y - 1);
   end Log;

   -------------------------------------------------------------------

   function Vectoint(X: Poly_Vector) return Integer is
      Y: Integer;
   begin
      Y := 0;
      for I in 0 .. M-1 loop
         if X(I) = 1 then 
            Y := Y + 2**I;
         end if;
      end loop;
      return Y;
   end Vectoint;      
   
   -------------------------------------------------------------------   
   
   function Inttovect(X: Integer) return Poly_Vector is
      R: Poly_Vector;
      I,Kr,Kc: Integer;
   begin   
      for I in 0 .. M-1 loop 
         R(I) := 0;
      end loop; 
      I := 0;
      Kc := X;
      Kr := 0;
      while Kc /= 1 loop
         Kr := Kc rem 2;
         Kc := Kc/2;
         R(I) := Kr;    
         I := I + 1;
      end loop;
      R(I) := 1;
      return R;
   end Inttovect;

   -------------------------------------------------------------------   

   function Nb_Exp(A: Poly_Vector; E: Integer; H: H_Array; W: W_Array) return Poly_Vector is
      B,C,R: Poly_Vector;
      --      h: h_array;
      --      w: w_array;
      V: Integer;
   
   begin
      V := M/2;   
   
      R:= Inttovect(E);
   
      C := A;
      for I in 0 .. M-1 loop
         B(I) := 1;
      end loop;
      for I in 0 .. M-1 loop
         if R(I) = 1 then
            B := Nb_Multiplier(B,C,H,W);
         end if;
         C := Nb_Sq(C);
      end loop;
      
      return B;    
      
   end Nb_Exp;

   ---------------------------------------------

   function Vectoint_K(X: Poly_Vector; K, P: Integer) return Integer is
      Y,J: Integer;
   begin
      Y := 0;
      J := 0;
      for I in P .. P+K-1 loop
         if X(I) = 1 then 
            Y := Y + 2**J;
         end if;
         J := J + 1;
      end loop;
      return Y;
   end Vectoint_K;   

   ----------------------------------------------------

   function Pow2j(Q: Integer) return Integer is
      J, P: Integer;
   begin
      P := 0;
      J := 1;
      while Q >= 2**J loop
         if Q = 2**J then P := J; end if;
         J := J + 1;
      end loop;
      return P;
   end Pow2j;


end Gf2m;

