----------------------------------------------------------------------
-- Galois Field Chapter 8 package (GF2m.ads)
-- 
-- Defines useful funtions for GF (2^m)
-- of Chapter 8 examples
---------------------------------------------------------------------
with Gnat.Io; use Gnat.Io;
with Finite_Fields_Gf2m; use Finite_Fields_Gf2m;

package Gf2m is

   M: constant Natural := 5;
   --   k: constant natural := 2;
   
   type Poly_Vector is new Bit_Vector(0 .. M-1);
   type Poly_Vector2 is new Bit_Vector(0 .. M-2);
   type Poly_Vectorm is new Bit_Vector(0 .. M);
   type Poly2_Vector is new Bit_Vector(0 .. 2*M-2);
   type Poly_Matrix_M1m2 is array (0 .. M-1, 0 .. M-2) of Bit;
   type Poly_Matrix_M2m1 is array (0 .. M-2, 0 .. M-1) of Bit;
   type Poly_Matrix is array (0 .. M-1, 0 .. M-1) of Bit;
   type H_Array is array (1 .. M/2) of Integer;
   type W_Array is array (1 .. M/2, 1 .. M-1) of Integer;
   type Yij_Array is array (0 .. M-1, 1 .. M/2) of Bit;
   type Yiv_Array is array (0 .. M/2-1) of Bit;
   type K_Array is array (1 .. M/2-1) of Integer;
   type R_Array is array (0 .. M/2-1) of Integer;
   
   --   P: constant poly_vector := (1,0,0,1,0,1,0,1);
  
   function Product_Alpha_A(A,F: Poly_Vector) return Poly_Vector;
   function Product_Alpha_Am(A,F: Poly_Vectorm) return Poly_Vectorm;
   function Product_Alpha(F: Poly_Vector) return Poly_Vector;
   function M2abv(X: Bit; Y: Poly_Vector) return Poly_Vector;
   function M2xvv(X: Poly_Vector; Y: Poly_Vector) return Poly_Vector;
   function M2xvvm(X: Poly_Vectorm; Y: Poly_Vectorm) return Poly_Vectorm;
   function M2xvv2(X: Poly2_Vector; Y: Poly_Vector) return Poly2_Vector;
   function Lshift(X: Poly_Vector) return Poly_Vector;
   function Rshift(X: Poly_Vector) return Poly_Vector;
   function Lshiftm(X: Poly_Vectorm) return Poly_Vectorm;   
   function Rshiftm(X: Poly_Vectorm) return Poly_Vectorm;
   function Lshift2(X: Poly2_Vector) return Poly2_Vector;
   function Poly_Multiplication(A,B: Poly_Vector) return Poly2_Vector;
   function Reduction_Matrix_R (F: Poly_Vector) return Poly_Matrix_M1m2;
   function Matrix_P (F: Poly_Vector) return Poly_Matrix_M2m1;
   function Mastrovito_Matrix (A: Poly_Vector; P: Poly_Matrix_M2m1) return Poly_Matrix;
   function Vector_D (A,B: Poly_Vector) return Poly_Vector;
   function Vector_E (A,B: Poly_Vector) return Poly_Vector2;
   function Lsbfirst_Squarer(A,F: Poly_Vector) return Poly_Vector;
   function Lsbfirst(A,B,F: Poly_Vector) return Poly_Vector;
   function Bsquarer_Montgomery(A,F: Poly_Vector) return Poly_Vector;
   function Bmult_Montgomery(A,B,F: Poly_Vector) return Poly_Vector;
   function Unitym(A: Poly_Vectorm) return Integer;
   function Degreem(A: Poly_Vectorm) return Integer;
   function Nb_Sq(A: Poly_Vector) return Poly_Vector;
   function H_Function_Gf2_4(A,B: Poly_Vector) return Bit;
   function Nb_Multiplier(A,B: Poly_Vector; H: H_Array; W: W_Array) return Poly_Vector;
   function Log(X: Integer) return Integer;
   function Vectoint(X: Poly_Vector) return Integer;
   function Inttovect(X: Integer) return Poly_Vector;
   function Nb_Exp(A: Poly_Vector; E: Integer; H: H_Array; W: W_Array) return Poly_Vector;
   function Vectoint_K(X: Poly_Vector; K,P: Integer) return Integer;
   function Pow2j(Q: Integer) return Integer;

end Gf2m;

