----------------------------------------------------------------------------
-- Type-I Optimal Normal Basis Multiplier (NB_T1_multiplier.vhd)
--
-- Computes the multiplication over normal basis in GF(2^m)
-- The hardware is generated for a specific f.
--
-- 
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package NB_T1_multiplier_parameters is
--------------------------------------------------
--  Constants for GF(2^4) 
--  constant M: integer := 4;
--  constant V: integer := 2;  -- V = M/2
--------------------------------------------------
--  Constants for GF(2^10) 
  constant M: integer := 10;
  constant V: integer := 5;  -- V = M/2
--------------------------------------------------
--  Constants for GF(2^28) 
--  constant M: integer := 28;
--  constant V: integer := 14;  -- V = M/2
--------------------------------------------------
--  Constants for GF(2^36) 
--  constant M: integer := 36;
--  constant V: integer := 18;  -- V = M/2
--------------------------------------------------
--  Constants for GF(2^58) 
--  constant M: integer := 58;
--  constant V: integer := 29;  -- V = M/2
--------------------------------------------------
--  Constants for GF(2^66) 
--  constant M: integer := 66;
--  constant V: integer := 33;  -- V = M/2
--------------------------------------------------
  type yij_array is array (0 to M-1) of std_logic_vector(1 to M/2);
  type k_array is array (1 to M/2-1) of integer;
  type yiv_array is array (0 to M/2-1) of std_logic;
--------------------------------------------------
--  Values for GF(2^4) 
--  constant k: k_array := (3);
--------------------------------------------------
--  Values for GF(2^10)
  constant k: k_array := (8,4,6,9);
--------------------------------------------------
--  Values for GF(2^28) 
--  constant k: k_array := (5,22,10,21,2,12,18,16,24,23,9,3,27);
--------------------------------------------------
--  Values for GF(2^36) 
--  constant k: k_array := (26,23,16,7,20,34,17,19,5,12,33,6,4,9,29,24,35);
--------------------------------------------------
--  Values for GF(2^58) 
--  constant k: k_array := (50,6,42,40,17,51,25,10,14,26,33,46,32,11,53,23,41,3,56,28,30,35,54,44,52,15,48,57);
--------------------------------------------------
--  Values for GF(2^66) 
--  constant k: k_array := (39,15,12,64,32,34,48,26,61,17,58,16,10,22,40,59,62,53,24,51,57,11,3,54,44,49,29,42,50,20,37,65);
--------------------------------------------------
end NB_T1_multiplier_parameters;

------------------------------------------------------------
-- NB_T1_multiplier
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.NB_T1_multiplier_parameters.all;

entity NB_T1_multiplier is
port (
  a, b: in std_logic_vector(M-1 downto 0);
  c: out std_logic_vector(M-1 downto 0)
  );
end NB_T1_multiplier;

architecture simple of NB_T1_multiplier is
  signal yij_s: yij_array;
  signal yiv_s: yiv_array;
  signal c_s, c_aux: std_logic_vector(M-1 downto 0);  
begin

  yij_process: process(a,b)
    variable yij: yij_array;
  begin
    for i in 0 to m-1 loop
      for j in 1 to v loop
        yij(i)(j) := (a(i) xor a((i+j) mod m)) and (b(i) xor b((i+j) mod m));
      end loop;
    end loop;
    yij_s <= yij;
  end process;

  yiv_process: process(a,b)
    variable yiv: yiv_array;
  begin
    for i in 0 to v-1 loop
      yiv(i) := (a(i) xor a((v+i) mod m)) and (b(i) xor b((v+i) mod m));
    end loop;
    yiv_s <= yiv;
  end process;

  c_s_process: process(a,b)
    variable caux: std_logic_vector(M-1 downto 0);
  begin
    for i in 0 to m-1 loop
      caux(i) := a(i) and b(i);
    end loop;
    c_s <= caux;
  end process;

  P1: process(yiv_s, yij_s, c_s)
    variable f: std_logic;
    variable r,r2,c_v: std_logic_vector(M-1 downto 0);
  begin
    f := yiv_s(0);
    c_v := c_s;
    for j in 1 to v-1 loop
      for i in 0 to m-1 loop
        r(i) := yij_s(i)(j);
      end loop;
      for i in 1 to k(j) loop 
        -- Squaring
        r2(0) := r(m-1);
        for i in 1 to m-1 loop
          r2(i) := r(i-1);
        end loop;
        r := r2;
      end loop;
      for i in 0 to m-1 loop
        c_v(i) := c_v(i) xor r(i);
      end loop;
      f := f xor yiv_s(j);
    end loop;
    if f = '1' then
      for i in 0 to m-1 loop
        c_v(i) := not c_v(i);
      end loop;
    end if;   
    c_aux <= c_v;
  end process;

   c <= c_aux;

end simple;
